#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#		A L E X A   H U E   B R I D G E   R E M O T E
#		version 3.0.1 for running remotely on the Raspberry Pi and connecting to XTension
#		https://machomeautomation.com/doku.php/supported_hardware/alexapi
#		copyright 2020 james@sentman.com
#
#
#	this version is built to run on a remote raspberry pi on the same network so that it can pretend better to
#	look like a hub to the now broken Amazon software
#	must be run as sudo so must be run as a user that can sudo without a password
#	must have access to port 80 so no other process can be running on port 80
#
#
#
#	Originally by Dean Davis of afterten.com now maintained by james@sentman.com
#	XTension Plugin version 2.1.1 http://MacHomeAutomation.com/
#
# pretends to be a hue hub in order that we can share devices to Alexa natively without
# requiring a cloud connection.
#
# version 1.0
#	there was no version 1.0, I considered that to be Deans original app and that makes this 2.0

# version 2.0.1 10/4/2018
#	added the local IP address sent from XTension into the settings object for the plugins
#	and if that is present and not 127.0.0.1 then we will use that for advertising our
#	services.

# version 2.0.2 10/5/2018
#	added a command to force the IP address that is advertised via the forceIp( whatever) script command
#	no interface to this but anyone who has this problem can use that command
#

# version 2.0.3 10/31/18
#	added extra error checking around the received UDP command handler as that seems to be erroring out
#	causing things to stop responding to alexa
#

# version 2.1 11/24/18
#	fixed a problem with the web server where it was basically single threaded and therefore any hung up
#	connections would get stuck and render it unable to serve any new connections.

# version 2.1.1 7/27/19
#	fixed a typo that caused an error in the error handling if the server handling of a request
#	failed for some lower level system reason

# version 3.0 2/20/2020 now can run as root and get around the limitations or purposeful changes made by Amazon
#

# version 3.0.1 3/1/2020
#	fixed problem sharing scripts to Alexa with the new cruelness


#
#	firmware that is broken version is: 644580520 on my echo dot.
#	trying to force it to update to something new now
#	though they never did fix the problem, so taking the matter into our own hands but still
# 	not starting our own cloud. That will come someday perhaps

import sys
import threading
import socket
import socketserver
from http import server
import time
import uuid
import os
import json
import struct
import random
import signal
import traceback
import math




#
# Importing Necessary XTension Connection Modules
#
from xtension_plugin.xtension_constants import *
from xtension_plugin.xtension_plugin import *



#
# plugin begins here
#

pluginVersion = '3.0.1pi' # used for logging and also inserted into the UPNP packets though that probably doesn't matter

# key for storing the light index in each shared unit
# this has to not change or things go goofy in the alexa and it forgets which unit was which
# we store the next available in our own settings and also the number that is assigned to a unit
# in the data of any unit that we share. That is handled in the xtUnit subclass

keyLastUsedIndex	= 'xt.alexa.lastusedindex'

# this is the key that goes in the unit however it is necessary to add the
# unique id of this interface to the end of it so that if the same unit or script is
# shared through more than one interface or was previously shared in another interface
# that we can tell them apart and not screw up all the sharing
keySharingIndex		= 'xt.alexa.sharingindex'

# force the server to advertise on a specific IP for complex networking environments and such
keyForceIp			= 'xt.alexa.forceip'
kForceIpCommand		= 'forceip'



#
# interface configuration keys
#

keyHubName		= 'xt.alexa.hubname'



#
# other settings keys
#

keyHubUUID			= 'xt.alexa.hubuuid'

UPNPPort			= 1900
UPNPBroadcastGroup 	= '239.255.255.250'

#
# state object keys for parsing data from the state command
#
keyState 			= 'state'
keyStateOn			= 'on'
keyStateBrightness	= 'bri'
keyStateHue			= 'hue'
keyStateSat			= 'sat'
keyStateXY			= 'xy'
keyStateColorTemp	= 'ct'
keyStateReachable	= 'reachable'

keyStateAlert		= 'alert'
kAlertNone			= 'none'
kAlertSelect		= 'select'
kAlertLongSelect	= 'lselect'

keyStateEffect		= 'effect'
kEffectNone			= 'none'
kEffectColorLoop	= 'colorloop'

keyStateType			= 'type'
kTypeExtendedColorLight	= 'Extended color light'
kTypeColorLight			= 'Color light'
kTypeColorTempLight		= 'Color temperature light'
kTypeDimmableLight		= 'Dimmable light'


#
# index of shared units by their phillips hue device index
#

sharedUnitIndex		= {}

#
#	in debug mode if you turn this off you'll get a LOT of logging of every UPNP packet on the network
#	useful for debugging but not helpful for higher level issues or debugging.
suppressUDPLogging	= True



#
#	G E T   N E X T   H U E   I N D E X
#
#	gets the next available hue index 
lastHueIndex = -1

def getNextHueIndex():
	# if we haven't initialized the global variable then get it here

	global lastHueIndex
	if lastHueIndex == -1:
		lastHueIndex = int( XTension.settings.get( keyLastUsedIndex, "0"))
	else:
		lastHueIndex += 1
		
	XTension.setXTensionData( keyLastUsedIndex, lastHueIndex)

		
	return lastHueIndex
		
#
#		M A K E   N E W   S C R I P T
#
def makeNewScript( data):
	return xtSharedScript( data)
	

#
#		S H A R E D   S C R I P T
#
class xtSharedScript( xtScript):
	def __init__(self, theData):
		xtScript.__init__( self, theData)

		
		# we're pretending to be a unit since the Alexa doesn't do scenes or scripts or anything else
		# so we need local state data to send it to make it think that it's done it's thing
		
		self.state 			= False
		self.brightness 	= 0
	
		global lastHueIndex
		
		
		self.sharingIndex = -1
# 		sharingIndexKey = XTension.settings.values[ xtUnitKeyUniqueId] + keySharingIndex
		sharingIndexKey = XTension.settings.get( xtUnitKeyUniqueId) + keySharingIndex
		
		if not theData.exists( sharingIndexKey):
			self.sharingIndex = getNextHueIndex()
			self.setXTensionData( sharingIndexKey, self.sharingIndex)
			
			XTension.debugLog( "no sharing index in script '" + self.name + "' setting to " + str( self.sharingIndex))
			
		else:
			self.sharingIndex = int( theData.get( sharingIndexKey))
			if self.sharingIndex > lastHueIndex:
				lastHueIndex = self.sharingIndex
				# save the new value into XTension so we can access it next launch or when the next unit changes
				XTension.setXTensionData( keyLastUsedIndex, lastHueIndex)
			
			XTension.debugLog( "sharing index for " + self.name + " is set to " + str( self.sharingIndex))

		# some idiot checking to let us know if something else is claiming that it already has the
		# same sharing address... This should not happen
		
		if self.sharingIndex in sharedUnitIndex and sharedUnitIndex[ self.sharingIndex] != self:
			self.writeLog( "sharing index collision, index " + str( self.sharingIndex) + " is already in use", xtLogRed)
			self.sharingIndex = getNextHueIndex()
			self.setXTensionData( sharingIndexKey, self.sharingIndex)
			self.writeLog( "new index (" + str( self.sharingIndex) + ") is applied, but Alexa may now be confused, consider rebuilding shared objects", xtLogRed)

		sharedUnitIndex[ self.sharingIndex] = self
		
	#
	#	S C R I P T   D E L E T E D
	#
	def scriptRemoved( self):
		XTension.debugLog( XTension.interfaceName + " is no longer sharing the script '" + self.name + "'")
		
		if self.sharingIndex in sharedUnitIndex:
			del sharedUnitIndex[ self.sharingIndex]
	#
	#   M A K E   U N I Q U E   I D
	#
	#	we need a MAC address style unique ID to insert into the json as a test to see if that is
	#	why the new alexa interface doesn't like this. We will use the unit's real unique ID and take the
	#	last 8 hex digits and stretch them out to look like a real one. If there arent 8 digits then
	#	we will just add 0's to pad
	
	def makeUniqueId( self):
		xtID = xtInt( self.data.get( xtUnitKeyUniqueId))
		
		s = hex( xtID).replace( '0x', '').replace( 'L', '')[::-1] # also remove the 0x at the front and reverse it
		out = []
		
		for i in range( 8):
			try:
				out.append( '0' + s[ i])
			except:
				# might not be enough digits, just add 0's in that case
				out.append( '00')
		theId = ':'.join( out)
		theId += '-0b' # probably not necessary but makes it look more like the real unique structure in the pi
		return theId
		

		
	def makeData( self):
	
		rootData = {}
		rootData[ 'name'] = self.name
		rootData[ 'modelid'] = 'XTENLIGHT001'
		rootData[ 'swversion'] = '5.23.1.13452'
		rootData[ 'uniqueid'] = self.makeUniqueId()
		rootData[ 'type'] = 'Dimmable light'
		
		
		stateData = {}

		stateData[ 'on'] = self.state
		stateData[ 'bri'] = self.brightness
		stateData[ 'alert'] = 'none'
		stateData[ 'reachable'] = True
# 		stateData[ 'mode'] = 'homeautomation'
		stateData[ 'effect'] = 'none'
		
		rootData[ 'state'] = stateData		
		
		rootData[ 'capabilities'] = { 'certified':True, 'streaming':{'renderer':False, 'proxy':False}}
		
		return rootData
		
		
	#
	#		S E T   S T A T E
	#
		
	def setState( self, rawData):
	
# 		XTension.debugLog( self.name + ": received state command (" + rawData + ")")
		XTension.debugLog( "%s: receved state command (%s)" % (self.name, rawData))
	
		try:
			jsonObject = json.loads( rawData)
		except Exception as e:
			XTension.writeLog( "setState json failed to parse " + str( e), xtLogRed)
			XTension.debugLog( rawData)
			return { "error":{"type":2, "address":"/lights/" + str( self.sharingIndex), "description":"invalid json"}}
			
		response = []
		
		
		if keyStateOn in jsonObject:
			# if it's on and we also have a non-zero brightness then this is a dim command
			# if it is on and we have no, or a zero brightness then it's just an on command
			# if it's off then it's an off command
			
			self.state = jsonObject[ keyStateOn]
			response.append( {"success":{"/lights/" + str( self.sharingIndex) + "/on":self.state}})
			
			# if the command is on then we run the script and also try to send a handler for the on turnon() handler
			# if the command is off then we don't run the script but we do try to run the on turnoff handler
			
			if self.state:
				self.execute()
				self.executeHandler( "commandon")
			else:
				self.executeHandler( "commandoff")

		
		if keyStateBrightness in jsonObject:
			self.brightness = jsonObject[ keyStateBrightness]
			response.append( {"success":{"/lights/" + str( self.sharingIndex) + "/bri":self.brightness}})
			
			# try to run the on setValue( newValue) handler in the script
			
			self.executeHandler( "commanddim", [ str( self.brightness)])
			

		return( response)
	
		
		

	


#
#		M A K E   N E W   U N I T
#
# returns an instance of our xtSharedUnit subclass to XTension when a 
# unit that is being shared to us is received
#
def makeNewUnit( data):	
	return xtSharedUnit( data)
	
	
	

#
#		S H A R E D   U N I T
#
# subclass of xtUnit to manage the sharing indexes and address
class xtSharedUnit( xtUnit):
	def __init__(self, theData):
		
		global lastHueIndex
		
		xtUnit.__init__( self, theData)
		
		#
		#	these are needed to properly return the last value to Alexa since the map and rounding does not result in the
		#	proper value being sent always and she responds with an error.
		# 	if the last value that we calculated is the same as whats in our database then we return the last unmapped
		#	brightness value we saved into these. If the real brightness is different then we re-calculate it as it changed
		#	elsewhere.
		
		self.lastSetBrightness = -1
		self.lastUnmappedBrightness = -1
		
		
		#self.data.subscribeToAll( self.dataChanged)
		
		
		# manage our sharing index
		# this is the bulb number or address as they are presented
		# in the Hue. This must be unique so to avoid problems when
		# adding or changing what we are sharing we will only ever increment
		# them for any given interface. 
		# if the unit has one stored already in it's data structure then
		# we will use that. Otherwise we will assign one from the last known
		# hue index as stored in the settings data for the interface. each
		# interface will have it's own index even if the unit is shared across more
		# than one for some instance. Each new interface that is created will store a
		# separate index into the unit, possibly creating some data structure bloat
		# if you delete and re-create this a dozen times
		#
		self.sharingIndex = -1
			
		sharingIndexKey = XTension.settings.get( xtUnitKeyUniqueId) + keySharingIndex

		if not theData.exists( sharingIndexKey):
			self.sharingIndex = getNextHueIndex()
			self.setXTensionData( sharingIndexKey, self.sharingIndex)
		
			self.debugLog( "no sharing index in '" + self.name + "' so using new value " + str( self.sharingIndex))
		else:
			self.sharingIndex = int( theData.get( sharingIndexKey))
			
			if self.sharingIndex > lastHueIndex:
				lastHueIndex = self.sharingIndex
				XTension.setXTensionData( keyLastUsedIndex, lastHueIndex)
				
			self.debugLog( "sharing index in unit '" + self.name + "' is: " + str( self.sharingIndex))
				
		# some idiot checking to let us know if something else is claiming that it already has the
		# same sharing address... This should not happen
		
		if self.sharingIndex in sharedUnitIndex and sharedUnitIndex[ self.sharingIndex] != self:
			self.writeLog( "sharing index collision, index " + str( self.sharingIndex) + " is already in use", xtLogRed)
			self.sharingIndex = getNextHueIndex()
			self.setXTensionData( sharingIndexKey, self.sharingIndex)
			self.writeLog( "new index (" + str( self.sharingIndex) + ") is applied, but Alexa may now be confused, consider rebuilding shared objects", xtLogRed)
		
		sharedUnitIndex[ self.sharingIndex] = self


	#
	#   M A K E   U N I Q U E   I D
	#
	#	we need a MAC address style unique ID to insert into the json as a test to see if that is
	#	why the new alexa interface doesn't like this. We will use the unit's real unique ID and take the
	#	last 8 hex digits and stretch them out to look like a real one. If there arent 8 digits then
	#	we will just add 0's to pad
	
	def makeUniqueId( self):
		xtID = xtInt( self.data.get( xtUnitKeyUniqueId))
		
		s = hex( xtID).replace( '0x', '').replace( 'L', '')[::-1] # also remove the 0x at the front and reverse it
		out = []
		
		for i in range( 8):
			try:
				out.append( '0' + s[ i])
			except:
				# might not be enough digits, just add 0's in that case
				out.append( '00')
		theId = ':'.join( out)
		theId += '-0b' # probably not necessary but makes it look more like the real unique structure in the pi
		return theId
		
		
		
	
	#
	#	M A K E   D A T A
	#
	# called to create the data structure that can be sent to the requestor
	# this does not include the index/address that is necessary to send all
	# the lights, that should be added before all the lights are concatenated together
	# and turned into JSON
	#
	# {"0":
	#	{"name":"shared pseudo 1","modelid":"XTENLIGHT001","swversion":"00000001","type":"Dimmable light",
	#		"state":{"on":false,"bri":0,"reachable":true,"alert":"none"}},
	
	

	def makeData( self):
	
		#self.data._debugLog()

		rootData = {}
		
		rootData[ 'name'] = self.get( xtUnitKeyName)
		rootData[ 'modelid'] = 'XTENLIGHT001'
		rootData[ 'swversion'] = '5.23.1.13452' # trying a real version number as well
		rootData[ 'uniqueid'] = self.makeUniqueId()
		
		# try the color data too
		# turns out the local connection of the hue in Alexa does not support
		# color anyway, but keep this in case that changes in the future
		#
		if self.get( xtUnitKeyAllowColor, False):
			
			if self.get( xtUnitKeyAllowColorTemp, False):
				# if both color temp and color then we are an extended color light
				rootData[ keyStateType] = kTypeExtendedColorLight #'Extended color light'
			else:
				# only color, but not colorTemp, we're a color light
				rootData[ keyStateType] = kTypeColorLight #'Color light'
		elif self.get( xtUnitKeyAllowColorTemp, False):
			# just color temperature we are a color temperature light
			rootData[ keyStateType] = kTypeColorTempLight #'Color temperature light'
		else:		
			# we are just a regular light only dimmable which turns out to be
			# the only kind that Alexa supports directly anyway
			rootData[ keyStateType] = kTypeDimmableLight #'Dimmable light'
		
		stateData = {}

		stateData[ keyStateOn] = self.getStatus()
		# brightness is from 0 to 255 so do the necessary work to map XTension's 
		# min/max values into a range of 0 to 255
		
		# seems alexa only supports color, enhanced color and dimmable lights, no support for on/off types
		# indeed it will ignore the device entirely if it's not a dimmable type
		
		if self.get( xtUnitKeyUseMinMax):
			unitMin = xtFloat( self.get( xtUnitKeyMinValue))
			unitMax = xtFloat( self.get( xtUnitKeyMaxValue))
		else:
			unitMin = 0
			unitMax = 100
		
		# we must return the same value that Alexa set us to or we she will tell the user that the unit
		# is not responding. Since it seems problematic to do the mapping in reserve we have saved off the last
		# unmapped value as well and if the current brightness in XTension has not changed we just always return that
		# if it has changed then we return a mapped value which is fine. It is only the checkback she makes immediately
		# after controlling the unit that must return the same value.
		
		currentValue = self.getValue()
		if currentValue != self.lastSetBrightness:
			thisValue = map( self.getValue(), unitMin, unitMax, 0, 254)
		else:
			thisValue = self.lastUnmappedBrightness
		
		stateData[ keyStateBrightness] = math.ceil( thisValue)

			

		stateData[ keyStateAlert] = kAlertNone #'none'
		
		# if there is no error for the unit then we are reachable
		
		if int(self.get( xtUnitKeyErrorLevel, 0)) != 0:
			stateData[ keyStateReachable] = False
		else:
			stateData[ keyStateReachable] = True
			
		stateData[ keyStateEffect] = kEffectNone #'none'

		# put the state data into the rootData JSON		
		rootData[ keyState] = stateData
		
		# add the other standard root data structures, probably not necessary but added for
		# complete identicality (I just made up that word I think) with the real Hue hub output
		rootData[ 'capabilities'] = { 'certified':True, 'streaming':{'renderer':False, 'proxy':False}}
		
		return rootData
		
		
	#
	#	S E T   S T A T E
	#
	# called when the http server receives a PUT state change command for this light
	# will receive the raw data which should be placed into an object via the json system
	# response should be appropriate given the input and should be an object
	# which will get turned into raw JSON after it is returned
	#
	
	def setState( self, rawData):
	
		XTension.debugLog( self.name + ": received state command (%s)" % rawData)
	
		try:
			jsonObject = json.loads( rawData)
		except Exception as e:
			XTension.writeLog( "setState json failed to parse " + str( e))
			XTension.debugLog( rawData)
			return { "error":{"type":2, "address":"/lights/" + str( self.sharingIndex), "description":"invalid json"}}
			
		response = []
		
		
		if keyStateOn in jsonObject:
			# if it's on and we also have a non-zero brightness then this is a dim command
			# if it is on and we have no, or a zero brightness then it's just an on command
			# if it's off then it's an off command
			
			newOnState = jsonObject[ keyStateOn]
			
				
			
			if not newOnState:
				self.turnOff()
				# our state data from XTension does not update fast enough as Alexa requests this
				# info immediately again after controlling us to see if it worked
				# so we have to manually update our data
				self.data.set( xtUnitKeyValue, 0)
				
				response.append( {"success":{"/lights/" + str( self.sharingIndex) + "/on":False}})
			else:
				if not keyStateBrightness in jsonObject or jsonObject[ keyStateBrightness] == 0:
					# this is a simple on and not a dim command
					self.turnOn()
					response.append( {"success":{"/lights/" + str( self.sharingIndex) + "/on":True}})
					
					if self.data.get( xtUnitKeyDimmable):
						workValue = xtFloat( self.data.get( xtUnitKeyPresetLevel, 1))
						if workValue == 0:
							workValue = 1
					else:
						workValue = 1
					
					self.data.set( xtUnitKeyValue, workValue)
					
				elif keyStateBrightness in jsonObject:
				
					if self.get( xtUnitKeyUseMinMax):
						unitMin = xtFloat( self.get( xtUnitKeyMinValue))
						unitMax = xtFloat( self.get( xtUnitKeyMaxValue))
					else:
						unitMin = 0
						unitMax = 100
				
					thisValue = int( map( jsonObject[ keyStateBrightness], 0, 254, unitMin, unitMax))
					
					# save these so we can return them in the next call to make data when Alexa checks up to see if we
					# really did set the value or not. Mapping between 0-100 amd 0-254 makes this difficult so just save them off
					self.lastSetBrightness = thisValue
					self.lastUnmappedBrightness = jsonObject[ keyStateBrightness]
					

					
					self.setValue( thisValue)
					self.data.set( xtUnitKeyValue, thisValue)
					self.setPreset( thisValue) # also set the preset here so that it will return to it the next on command
					
					response.append( {"success":{"/lights/" + str( self.sharingIndex) + "/bri":jsonObject[ keyStateBrightness]}})
					
		elif keyStateBrightness in jsonObject:
		
			newValue = jsonObject[ keyStateBrightness]
			if newValue == 0:
				self.turnOff()
				self.data.set( xtUnitKeyValue, 0)
				
				response.append( {"success":{"/lights/" + str( self.sharingIndex) + "/bri":0}})
			else:
				if self.get( xtUnitKeyUseMinMax):
					unitMin = xtFloat( self.get( xtUnitKeyMinValue))
					unitMax = xtFloat( self.get( xtUnitKeyMaxValue))
				else:
					unitMin = 0
					unitMax = 100
				
				thisValue = int( map( jsonObject[ keyStateBrightness], 0, 254, unitMin, unitMax))

				# save these so we can return them in the next call to make data when Alexa checks up to see if we
				# really did set the value or not. Mapping between 0-100 amd 0-254 makes this difficult so just save them off
				self.lastSetBrightness = thisValue
				self.lastUnmappedBrightness = jsonObject[ keyStateBrightness]

				self.setValue( thisValue)
				self.data.set( xtUnitKeyValue, thisValue)
				self.setPreset( thisValue) # also set the preset here so that it will return to it the next on command
				
				response.append( {"success":{"/lights/" + str( self.sharingIndex) + "/bri":newValue}})
				
		return( response)
			
	#
	#		U N I T   R E M O V E D
	#
	#	we get this event from the superclass when this unit is deleted, no longer assigned to our interface
	#	or when it's no longer being shared to us.
	#	note that our ID's and other indexes have already been removed at this point
	#
	def unitRemoved( self):
		if self.sharingIndex in sharedUnitIndex:
			del sharedUnitIndex[ self.sharingIndex]
			XTension.debugLog( "no longer sharing unit '" + self.name + "'")
		else:
			XTension.debugLog( "request to remove " + self.name + " but not in the sharing index")
		
		








#
#
#	class 		H U E   B R I D G E
#
#	all the work to manage the connections and servers is done inside this base class
#	



class hueBridge():

	#
	#	I N I T
	#
	def __init__( self):
	
		# create a holder for our HTTP server 
		self.httpManager = None

	
		# we require the UUID and hub name in order to start broadcasting
		# and it must be the same every time we startup or Alexa will think we are a different hub
		# if it's not already set in our XTension settings data object then we need to create one
		if not XTension.settings.exists( keyHubUUID):
			self.hubUUID = str( uuid.uuid1())
			XTension.setXTensionData( keyHubUUID, self.hubUUID)
		else:
			self.hubUUID = XTension.settings.get( keyHubUUID)
			
		self.hubName = XTension.settings.get( keyHubName, "XTension")
		
		# this is used to locate our resources directory so we can load them
		# by pluginDirectory I mean our current directory that this plugin is running from
		# even though this is not a plugin since it's running remotely it is still the same
		self.pluginDirectory = os.path.dirname( os.path.realpath( __file__))

		# make our sockets for sending the keep alive messages and responses to M-Search packets
		self.broadcastSocket = socket.socket( socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
		self.broadcastSocket.setsockopt( socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 20)
		self.responseSocket = socket.socket( socket.AF_INET, socket.SOCK_DGRAM)


		# load up the things we will need to serve from the resources directory
		# like the description file and the icons
		self.loadResources() 



		# get our local IP address.
		# we need this so that we can insert it into our broadcast advertisement packets
		# so that any interested client will be able to connect to us.
		
		# add this command handler to force the IP address that is advertised
		# this may not be necessary on the raspberry pi as it can be on the Mac but people
		# do all sorts of interesting things network config wise so keep it in for those special people
		XTension.addScriptHandler( kForceIpCommand, self.forceIp)
		
		# if there is a forceIp set in the XTension settings then honor that
		# otherwise use the getIP call
		# this ignores the local address that XTension places into the data as that would
		# not be correct for the Pi
		#
		self.serverIp = XTension.settings.get( keyForceIp, '')
		if self.serverIp != '':
			XTension.writeLog( "Alexa server forcing ip address to: " + self.serverIp)
		else:
			self.serverIp = getIp()
			
		XTension.debugLog( "local IP address is calculated as: " + self.serverIp)
		
		# the listening port is always port 80 now due to bugs in the new Alexa which
		# is why all this raspberry pi and running as root stuff is needed
		# we no longer get it out of the XTension settings, but just hard code to 80
		#self.serverPort = int( XTension.settings.get( xtKeyRemotePort, defaultServerIp))
		self.serverPort = 80 # always 80 now 
		
		
		
	def startup( self):
		#
		# startup the web server
		#
		self.httpManager = HTTPManager()
		
		# 
		#	register for the shutdown command from XTension so we can know if we are supposed to quit
		#
		XTension.addCommandHandler( xtCommandShutdown, self.shutdown)

		#
		#	startup the UPNP listening sockets
		#
		self.UPNPListener = socket.socket( socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
		self.UPNPListener.setsockopt( socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.UPNPListener.bind( ('0.0.0.0', UPNPPort))
		# add the broadcast group
		mreq = struct.pack( "4sl", socket.inet_aton( UPNPBroadcastGroup), socket.INADDR_ANY)
		self.UPNPListener.setsockopt( socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

		# socket must have a timeout so that the thread can wake up and see if we have received the shutdown command
		self.UPNPListener.settimeout( 1)

		# the listening for packets is done in a separate thread
		self.listenThread = Thread( target=self.threadedListener, args=())
		self.listenThread.start()

		
		# startup our broadcast keep alive timer
		
		self.broadcastTimer = xtBroadcastTimer( 60)
		self.broadcastTimer.start()

		# and finally tell the world we are here		
		self.sendBroadcastHello( True)
		
		
	#
	#	T H R E A D E D   L I S T E N E R
	#
	#	started as a thread in self.startup() this handles the reception of UPNP packets
	# 	until we have received the shutdown command, then it closes it's sockets and 
	# 	ends the thread so that we can quit.
	#
	def threadedListener( self):

		while not XTension.isShuttingDown:
			try:
				workPacket = self.UPNPListener.recvfrom( 8192)
			except socket.timeout:
				continue
				
			except:
				continue
				
			theData = workPacket[ 0]
			theHost = workPacket[ 1][ 0]
			thePort = workPacket[ 1][ 1]
			
			self.UPNPRequestReceived( theData, theHost, thePort)
			
		# have fallen out of the above check for isShuttingDown so close our socket
		# and return in order to end this thread.
		try:
			self.UPNPListener.close()
		except:
			pass
		
		
		
		
		
	
	#
	#	F O R C E   I P
	#
	#	the script call when you force the IP address by using the
	# 	tell xInterface "name of alexa interface" to forceIp( "192.168.0.4") 
	# 	or to clear:
	#	tell xInterface "name of alexa interface" to forceIp()
	#
	def forceIp( self, commandName, positionalParms, dataParms):
		try:
			if positionalParms == None:
				# in the Mac version we would check to see if XTension had told us what the local IP of that machine was
				# in this case we do not want to do that as we are running on a separate machine we just use getIp
				#if XTension.settings.exists( xtKeyLocalAddress):
				#	self.serverIp = XTension.settings.get( xtKeyLocalAddress)
				#	if self.serverIp == "127.0.0.1":
				#		self.serverIp = getIp()
				#else:
				#	self.serverIp = getIp()
				
				self.serverIp = getIp()
					
				XTension.setXTensionData( keyForceIp, '')
				
				XTension.writeLog( "Alexa server forced IP has been cleared, now advertising on: " + self.serverIp)
			else:
				
				self.serverIp = positionalParms[0]
				XTension.setXTensionData( keyForceIp, self.serverIp)	
				XTension.writeLog( "The Alexa server will advertise itself on IP: " + self.serverIp)
				
		except Exception as e:
			XTension.writeLog( "error while setting forced ip: " + str( e))
	
	
	
		
	#
	#		L O A D   R E S O U R C E S
	#
	#	these are the data structures that we need to serve up from the resources folder
	#
	def loadResources( self):
		
		try:

			# load description xml file
					
			workFile = open( os.path.join( self.pluginDirectory, "resources/description.xml"), 'r')
			self.rawDescription = workFile.read()
			# conform the line endings to make sure that it has crlf line endings
			self.rawDescription = '\r\n'.join( self.rawDescription.split( '\n'))
			workFile.close()
		
			# load icon data for serving the icon
			# note that in python3 we must open this as a binary 'bytes' file or it will refuse to load 
			# the binary data. See the 'rb' as the open parameter below
			workFile = open( os.path.join( self.pluginDirectory, "resources/icon.png"), 'rb')
			self.rawIcon = workFile.read()
			workFile.close()
			
			# load the upnp broadcast packet
			workFile = open( os.path.join( self.pluginDirectory, "resources/upnp_broadcast.txt"), 'r')
			workData = workFile.read()
			workFile.close()
			# conform the line endings to make sure that it has crlf line endings
			self.broadcastPacket = '\r\n'.join( workData.split( '\n'))
			
			# load the upnp response packet
			workFile = open( os.path.join( self.pluginDirectory, "resources/upnp_response.txt"), 'r')
			workData = workFile.read()
			workFile.close()
			# conform the line endings to make sure that it has crlf line endings
			self.responsePacket = '\r\n'.join( workData.split( '\n'))
			
			
			
			
			
		except Exception as e:
			XTension.writeLog( "error while loading resources, unable to continue: " + str( e), xtLogRed)
			XTension.setRunState( xtRunStateFail, "error while loading resources " + str( e))
			XTension.setStatus( str( e))
			sys.exit()
			
			
	
	

	
	

		
	
	
	#
	#		U P N P   R E Q U E S T   R E C E I V E D
	#
	# a packet has been received on the shared UPNP port in XTension and forwarded to us
	# 
	def UPNPRequestReceived( self, receivedData, theHost, thePort):
		
		# make sure that an error in here doesn't stop this thread
		# also added to the regular plugin handlers but that can't be distributed with 
		# this test versions so also add it here
		
		try:

			# received data will likely be a bytes like objectd but we need a string to parse it		
			if type( receivedData) != str:
				receivedData = receivedData.decode()

			# if our own hub ID is in the packet, then this is just an echo of our own broadcast or keep alive message
			# so there is no reason to process it at all			
			if self.hubUUID in receivedData:				
				return
		
			# if it's not an M-Search then also just ignore it, but log it if we are logging such things
			if not "M-SEARCH" in receivedData:
				if not suppressUDPLogging:
					XTension.debugLog( "%s:%s received non M-SEARCH packet: %s" % (theHost, thePort, receivedData))
					return
				else:
					return

				
			# if it is an M-Search then log it if we have logging turned on
			XTension.debugLog( "%s:%s received M-SEARCH packet" % (theHost, thePort))
			XTension.debugLog( receivedData)
			
			# only respond to the base item request, not anything else:
			# it seems that sometimes Alexa sends different discovery packets looking for the hub and we should respond to them too
			# filtering all but the basic message request seems to stop alexa discovery of the hub in some circumstances
			
			if "ST: urn:schemas-upnp-org:device:basic:1" in receivedData or "ST: upnp:rootdevice" in receivedData or "ST: ssdp:all" in receivedData:
		
				thisResponse = self.responsePacket % { "ipAddress":self.serverIp, "port":bridge.serverPort, "version":str( pluginVersion), "uuid":self.hubUUID}
		
				XTension.debugLog( "responding to M-SEARCH from %s:%s with: " % (theHost, thePort))
				if not suppressUDPLogging:
					XTension.debugLog( thisResponse)
	
# 				XTension.debugLog( "initial type of thisResponse: %s" % type( thisResponse))
				thisResponse = thisResponse.encode( 'utf-8')
# 				XTension.debugLog( "after encoding to bytes object: %s" % type( thisResponse))
	
				self.responseSocket.sendto( thisResponse, (theHost, thePort))
				
# 				XTension.debugLog( "back from sendto")
				
				self.setStatus( "Search Packet from %s" % theHost)
			else:
				XTension.debugLog( "not asking for basic device, not responding")
				
		except Exception as e:
			XTension.writeLog( "error while handling UPNP request: " + str( e), xtLogRed)
			XTension.writeLog( traceback.format_exc(), xtLogRed)

		
		
		
	#
	#		S E N D   B R O A D C A S T   H E L L O
	#
	#	must periodically send this to let devices know that we're still alive on the UPNP network
	# 	pass a false to send our going away message when we are told to shutdown
	#
	
	def sendBroadcastHello( self, isAlive):
	
		try:
			if isAlive:
				state = "alive"
			else:
				state = "byebye"
			
			workPacket = self.broadcastPacket % {"broadcastIp":UPNPBroadcastGroup, "upnpPort":str( UPNPPort),
												 "ipAddress":self.serverIp, "port":bridge.serverPort,
												 "version":pluginVersion, "state":state, "uuid":self.hubUUID}


			XTension.debugLog( "sending broadcast keep alive message")
			XTension.debugLog( workPacket)

			if self.broadcastSocket == None:
				self.broadcastSocket = socket.socket( socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
				self.broadcastSocket.setsockopt( socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 20)

			self.broadcastSocket.sendto( workPacket.encode( 'utf-8'), (UPNPBroadcastGroup, UPNPPort))
		except Exception as e:
			XTension.writeLog( 'error: bridge.sendBroadcastHello( %s)' % e)
		
											 

	#
	#		S H U T D O W N
	#
	#	called when XTension gets it's shutdown command, stop the broadcast, close sockets and close cleanly
	#
	def shutdown( self, theCommand):
		
		# stop the broadcast alive timer
		if self.broadcastTimer != None:
			self.broadcastTimer.stop()
			self.broadcastTimer = None
		
		# send the UPNP byby command
		self.sendBroadcastHello( False)
				
		# stop the http server

		if self.httpManager != None:
			self.httpManager.shutdown()


			
	#
	#		S E T   S T A T U S
	#
	#	sets the status in XTension temporarily to a value
	# 	and then returns it to the default in 5 seconds
	#
	
	def setStatus( self, theStatus):
		XTension.setStatus( theStatus)
		resetStatus.setTimeout( 5)
		
		
				


#
#		B R O A D C A S T   T I M E R
#
#	created to just send the keep alive message every minute
#
class xtBroadcastTimer( xtTimer):
	def action( self):
		if XTension.isShuttingDown:
			self.stop()
			return
			
		bridge.sendBroadcastHello( True)



#
#	class 	H T T P D   H A N D L E R
#
#	new class for handling the requests based on the python3 threading mixin server
#
class httpdHandler( server.BaseHTTPRequestHandler):

	#
	#	L O G   R E Q U E S T
	#
	# if we are in debug mode this logs to XTension all requests to the web server
	#
	
	def log_request( self, code):
		XTension.debugLog( code)
		
	#
	#	CHUNKED READING
	#
	#	this is not used by the Alexa process, but is needed when faking it out and testing it via CURL
	#	as it will do this which breaks everythign otherwise so support it
	#
	
	def readChunked( self):
		try:
			data = []
			while True:
				try:
					chunk_size = self.get_chunk_size()
					if chunk_size == 0:
						return ''.join( data)
					else:
						data.append( self.get_chunk_data( chunk_size))
					
				except Exception as e:
					XTension.writeLog( "alexa plugin error in readChunked: " + str( e))
					return ''.join( data)
		except Exception as e:
			XTension.writeLog( 'error readChunked( %s)' % e, xtLogRed)
			XTension.writeLog( traceback.format_exc(), xtLogRed)
			
	def get_chunk_size( self):
		
		try:
			size_str = self.rfile.read( 2)
			while size_str[-2:] != b"\r\n":
				size_str += self.rfile.read( 1)
			return int( size_str[:-2], 16)
		except Exception as e:
			XTension.writeLog( "alexa plugin error in get_chunk_size:" + str( e))
			XTension.writeLog( traceback.format_exc(), xtLogRed)
			return 0
			
		
	def get_chunk_data( self, chunk_size):
		
		try:
			data = self.rfile.read( chunk_size)
			self.rfile.read( 2)
			return data
		except Exception as e:
			XTension.writeLog( "alexa plugin error in get_chunk_data: " + str( e))
			XTension.writeLog( traceback.format_exc(), xtLogRed)
			return ""
	
	
	
	#
	#		D O    P O S T
	#
	#	this is how the alexa device asks us for a username to pass in the link later
	#	a JSON POST to /api containing {"devicetype":"Echo"} or some such
	#	we can ignore the specifics of that and just always return a success message which 
	# 	is something like this with a 40 character random username
	# [{"success":{"username":"zHM7rphKml0mECdJxnSM6l5lucRnRqXLHHUrs-lu"}}]
	#

	def do_POST( self):
		try:
			XTension.debugLog( "POST: " + self.path, xtLogGreen)
		
			if self.headers.get( 'Expect', '') == '100-continue':
				self.wfile.write( "100 Continue\r\n")
				theData = seld.readChunked()
				XTension.debugLog( "chunked data read (%s)" % theData)
			else:
				contentLen = int( self.headers.get( 'Content-Length', 1024))
				theData = self.rfile.read( contentLen)


			XTension.debugLog( " posted data (%s)" % theData)
		
			replyData = json.dumps( [ {"success":{"username":makeRandomUsername()}}]).encode( 'utf-8')

			self.send_response( 200)
			self.send_header( "Content-Type", "application/json")
			self.send_header( "Content-Length", len( replyData))
			self.end_headers()
			self.wfile.write( replyData)
		except Exception as e:
			XTension.writeLog( 'error do_POST( %s)' % e, xtLogRed)
			XTension.writeLog( traceback.format_exc(), xtLogRed)

	#
	#	D O   P U T
	#
	#	actual control of things ends up here
	#	this is the PUT method handler for the HTTP server
	#
	def do_PUT( self):
		try:
			(self.remoteAddress, self.remotePort) = self.client_address
			self.protocol_version = "HTTP/1.1"
			XTension.debugLog( "PUT: " + self.path, xtLogGreen)
		
			#
			# /api/<username>/lights/<id>/state
			#
		
			if self.headers.get( 'Expect', '') == "100-continue":
				self.wfile.write( "100 Continue\r\n")
				#XTension.writeLog( "continuing...")
				#sleep( 1)
			
				putData = self.readChunked()
				XTension.debugLog( putData)
			
			else:		
				contentLen = int( self.headers.get( 'Content-Length', 1024))
				putData = self.rfile.read( contentLen)
		
			parts = self.path.upper().split( '/')
			del parts[0]
		
			if len( parts) < 5:
				XTension.writeLog( "unsupported API PUT action")
				self.send_response( 400)
			
			if parts[ 2] == "LIGHTS" and parts[ 4] == "STATE":

				if not int( parts[ 3]) in sharedUnitIndex:
					XTension.writeLog( "attempted state control for unknown light at index: " + parts[ 3], xtLogRed)
					self.send_response( 404)
					return
				workUnit = sharedUnitIndex[ int( parts[3])]
				
				self.sendJSONData( workUnit.setState( putData))
				bridge.setStatus( "controlling " + workUnit.name)
				
			else:
				self.send_response( 400)
				XTension.writeLog( "unsupported API PUT command: " + self.path, xtLogRed)
		except Exception as e:
			XTension.writeLog( "alexa plugin error in do_PUT: " + str( e), xtlogRed)
			XTension.writeLog( traceback.format_exc(), xtLogRed)

			
	
	
	
	
	#
	#	D O   G E T
	#
	# GET requests end up here, getting the light status and such
	#

	def do_GET( self):
	
		try:

			XTension.debugLog( "HTTP REQUEST: " + self.path, xtLogGreen)
			self.remoteAddress = self.client_address[0]
			self.remotePort = self.client_address[1]
			self.protocol_version = "HTTP/1.1"
			

		
			# convert all the path parts to uppercase for the comparisons below
			# probably not necessary but eliminates another potential failure mode
			parts = self.path.upper().split( '/')

			# since the path always starts with a slash we can remove the first empty element
			# that the split command above will insert
			del parts[0]
		

			if len( parts) == 0:
				self.sendHello()
				return

			if parts[0] == '': #base request, redirect to index
				self.send_response( 301)
				self.send_header( "Location", "/index.html")
				self.end_headers()
				return
		
			# Send Description file		
			if parts[ 0] == "DESCRIPTION.XML":
				XTension.debugLog( "sending description XML to " + self.remoteAddress)
				self.sendDescription()
				bridge.setStatus( "sending xml to " + self.remoteAddress)
			
			# send logo, probably not necessary as this version of Alexa doesn't ask for it
			elif parts[ 0] == "HUE_LOGO_0.PNG" or parts[ 0] == "HUE_LOGO_3.PNG":
				XTension.debugLog( "sending logo to " + self.remoteAddress)
				self.sendIcon()
			
			# send index page, probably not necessary but does help us identify this server is working
			elif parts[ 0] == "INDEX.HTML" or parts[ 0] == "INDEX.HTM":
				XTension.debugLog( "request for index.html from " + self.remoteAddress)
				self.sendHello()
			
			# have received an API request
			elif parts[ 0] == "API":
				self.handleGetAPI( parts)
			
			else:
				# unknown request or not found or otherwise wrong
				XTension.debugLog( "unhandled hue bridge request: " + self.path)
				self.send_response( 400)
				
		except Exception as e:
			XTension.writeLog( "alexa plugin error in do_GET: " + str( e))
			XTension.writeLog( traceback.format_exc(), xtLogRed)

			
		
		


			
			
	#
	# 	H A N D L E   G E T   A P I
	#
	# /api/<username>/lights
	# /api/<username>/lights/3
	#
	def handleGetAPI( self, parts):
		XTension.debugLog( "api call: " + "/".join( parts))
		
		# send the DEEP SCAN results
		# probably not necessary but added for full indenticality with the hue hub
		if len( parts) == 2:
			self.sendDeepScan()
			return
		
		parts.append( '') # to keep from getting an error looking for the light index below
		
		APIMethod = parts[ 2]
		
		if APIMethod == "LIGHTS":

			# if no specific light index specified then we send all the lights
			if parts[ 3] == '':
				self.sendAllLights()
				
			else:
				# asking for a specific light make sure it exists and then send it			
				if int( parts[3]) in sharedUnitIndex:
					self.sendLightByIndex( int( parts[ 3]))
				else:
					XTension.debugLog( "request for unknown light index: " + parts[3])
					self.send_response( 404)

		else:
			XTension.debugLog( "unsupported API call: " + self.path)
			# but don't send an error, just send an empty json item
			self.send_response( 200)
			self.send_header( "Content-Type", "text/json")
			self.send_header( "Content-Length", "2")
			self.end_headers()
			self.wfile.write( "{}".encode( 'utf-8'))
	
	#
	#	S E N D   D E E P   S C A N
	#
	def sendDeepScan( self):
		self.send_response( 400)
		return
		
		root = {}
		lights = {}
		root[ "lights"] = lights
		myList = []
		
		for thisKey in sharedUnitIndex:
			myList.append( thisKey)
			
		myList.sort()
		
		for thisKey in sharedUnitIndex:
			lights[ str( thisKey)] = sharedUnitIndex[ thisKey].makeData()
			
		self.sendJSONData( root)
	
	#
	#	S E N D   A L L   L I G H T S
	#
	# updated to always send the lights in order as that is a difference between this code
	# and the actual original hue output.
	#
	def sendAllLights(self):
	
		root = {}
		sortedKeys = []
		
		for thisKey in sharedUnitIndex:
			sortedKeys.append( thisKey)
			
		sortedKeys.sort()
		
		out = ''
		
		for thisKey in sortedKeys:
			if out != '':
				out += ', '
			
			out += '"%s": ' % thisKey
			
			out += json.dumps( sharedUnitIndex[ thisKey].makeData())
			
		out = '{' + out + '}'
		out = out.encode( 'utf-8')
		
		self.send_response( 200)
		self.send_header( "Content-Type", "application/json")
		self.send_header( "Content-Length", str( len( out)))
		self.send_header( "Connection", "Close")
		self.end_headers()
		self.wfile.write( out)
		
			
# 			root[ str( thisKey)] = sharedUnitIndex[ thisKey].makeData()
			
# 		self.sendJSONData( root)
		
	
	
	#
	#	S E N D   L I G H T   B Y   I N D E X
	#
	
	def sendLightByIndex( self, index):
		if not index in sharedUnitIndex:
			self.send_response( 404)
			return
			
		self.sendJSONData( sharedUnitIndex[ index].makeData())
			
	
	
	#
	#	S E N D   J S O N   D A T A
	#
	# helper function so I don't have to keep typing this little block of code everywhere
	#
	def sendJSONData( self, data):
		jsonData = json.dumps( data)
		XTension.debugLog( "sending:" + jsonData)
		
		jsonData = jsonData.encode( 'utf-8')
		self.send_response( 200)
		self.send_header( "Content-Type", "application/json")
		self.send_header( "Content-Length", str( len( jsonData)))
		self.send_header( "Connection", "Close")
		self.end_headers()
		self.wfile.write( jsonData)
	
	
	
	#
	#	S E N D   H E L L O
	#
	# this isn't actually necessary but might help in debugging or testing to see if the
	# server is running
	#
	def sendHello( self):
		helloMessage = "XTension Phillips Hue Bridge v%s\nhttp://MacHomeAutomation.com/" % pluginVersion
		helloMessage = helloMessage.encode( 'utf-8')
		
		self.send_response( 200)
		self.send_header( "Content-Type", "text/plain")
		self.send_header( "Content-Length", str( len( helloMessage)))
		self.end_headers()
		
		self.wfile.write( helloMessage)
	
	#
	#	S E N D   I C O N
	#
	def sendIcon( self):
		
		self.send_response( 200)
		self.send_header( "Content-Type", "image/png")
		self.send_header( "Content-Length", str( len( bridge.rawIcon)))
		self.end_headers()
		
		self.wfile.write( bridge.rawIcon)
		
			
			
	#
	#	S E N D   D E S C R I P T I O N
	#
	# JFS this is another difference, we want to put the IP of this machine into here and not the serverIP as that is not correct when running remotely
	
	def sendDescription( self):
		try:
			thisDescription = bridge.rawDescription % { "ipAddress":bridge.serverIp,"port":bridge.serverPort,"uuid":bridge.hubUUID}
			thisDescription = thisDescription.encode( 'utf-8')
															
			self.send_response( 200)
		
			self.send_header( "Cache-Control", "no-store, no-cache, must-revalidate, post-check=0, pre-check=0")
			self.send_header( "Content-Type", "text/xml")
			self.send_header( "Content-Length", str( len( thisDescription)))
			self.send_header( "Connection", "Close")
			self.send_header( "Server", "nginx")
			self.end_headers()
		
			self.wfile.write( thisDescription)
		except Exception as e:
			XTension.writeLog( "error sendDescription( %s)" % e, xtLogRed)
			XTension.writeLog( traceback.format_exc(), xtLogRed)

		
		





#
#	class 	ThreadedServer
#
# parent class for the threaded mix-in web server in python3
#

class ThreadedServer( socketserver.ThreadingMixIn, server.HTTPServer):
	allow_reuse_address = True
	daemon_threads = True
	

#
#	class 	H T T P   M A N A G E R
#
#	class that manages the new properly threaded HTTP Server
#
class HTTPManager():
	def __init__( self):
		self.httpd 			= None
		self._thread 		= None
		self.shuttingDown 	= False
		
		self.startup()
		
	#
	#	S T A R T U P
	#
	#	called only from our init method so why bother being a separate call at all?
	#
		
	def startup( self):
		# this is not necessary in this case as we will never be restarting the server in the same
		# session. We start it at the beginning and stop it when we receive the shutdown command
		self.shutdown()
		self.shuttingDown = False
		
		try:
			self.httpd = ThreadedServer( ('0.0.0.0', bridge.serverPort), httpdHandler)
			self.httpd.server_bind = self.httpd.server_close = lambda self: None
		
			self._thread = Thread( target=self.serverThread, args=(), name='httpThread')
			self._thread.start()
		except Exception as e:
			XTension.writeLog( 'Error: HTTPManager.startup( %s)' % e)
		
	#
	#	S H U T D O W N
	#
	#	called by the bridge class when it receives the shutdown command from XTension
	# 	we stop our thread and close the server
	#
	def shutdown( self):
		if self.httpd == None:
			return
			
		self.shuttingDown = True
		try:
			self.httpd.shutdown()
		except:
			pass
			
		self._thread = None
		self.httpd = None
		
	#
	#	S E R V E R   T H R E A D
	#
	#	started as a thread in our startup() method and we server forever until the shutdown
	#	method is called
	#
	def serverThread( self):
		try:
			self.httpd.serve_forever()
		except Exception as e:
			if not self.shuttingDown:
				XTension.writeLog( 'Error in http server thread: %s' % e)
				
				self.http = None
				self._thread = None
		

		
		
		
#
#	L I S T   U N I T S
#
#	registered for the listUnits() script handler
# debugging aid to print out the data for the units that we think we are sharing
#
def listUnits( commandName, positionalParms, dataParms):
	
	for key in sharedUnitIndex:
		workUnit = sharedUnitIndex[ key]
		XTension.writeLog( str( key) + " = " + str( workUnit.makeData()))
		
		
			
			
def map( value, in_min, in_max, out_min, out_max):
	return (value - in_min) * (out_max - out_min) / (in_max - in_min) + out_min


#
# this is now used only if there is no valid IP address in the XTEnsion settings object
#
def getIp():
	s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	try:
		s.connect(('10.255.255.255', 1))
		IP = s.getsockname()[0]
	except Exception as e:
		#IP = '127.0.0.1'
		XTension.writeLog( "unable to resolve a local IP address, alexa bridge will not be available: " + str( e), xtLogRed)
		XTension.setRunState( xtRunStateFail, "unable to resolve a local address")
		XTension.setStatus( "unable to resolve network address")
		sys.exit()
        
	finally:
		s.close()
		
		# Michael got this when his ethernet cord was unplugged when he started up the interface
		if IP == "0.0.0.0":
			XTension.writeLog( 'unable to resolve a local IP address, please check network connections or use the forceIp( "x.x.x.x") command to set one manually', xtLogRed)
			XTension.setRunState( xtRunStateFail, "no IP address")
			sys.exit()
			
		return IP
#
# 	R E S E T   S T A T U S
#
# used to reset the status to it's normal state after we have changed it to indicate an action
#
class cResetStatus( XTDoLater):
	
	def action( self):
		XTension.setStatus( "Listening on port: %s" % bridge.serverPort)


#
#	M A K E   R A N D O M   U S E R N A M E
#
#
#	creates a 40 character random username just like the Hue does
#

def makeRandomUsername():
	s = "01234567890AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz"
	alen = len( s) - 1
	out = ''
	
	for i in range( 40):
		out += s[ random.randint( 0, alen)]
				
	return out
	
	
	
	
def signal_handler( sig, frame):
	XTension.writeLog( 'Shutting down for signal( %s)' % sig)
	
	if XTension != None:
		XTension.doShutdown()
	
	# this is PROBABLY not necessary as it will have already done so when the callback from the
	# XTension shutdown command above is called, but just to be sure idiot check it
	if bridge != None:
		bridge.shutdown()


def xtRound( x):
	justInt = float( int( x))
	fract = x - justInt

	if fract >= 0.5:
		XTension.debugLog( 'value=%s integer=%s fract=%s rounded=%s' % (x, justInt, fract, justInt+1))
		return justInt + 1
	else:
		XTension.debugLog( 'value=%s integer=%s fract=%s rounded=%s' % (x, justInt, fract, justInt))
		return justInt
	
	
	
	

#
#
#		M A I N
#
#

bridge = None
XTension = None

signal.signal( signal.SIGINT, signal_handler)
signal.signal( signal.SIGTERM, signal_handler)


XTension = cXTension()
XTension.onMakeNewUnit = makeNewUnit
XTension.onMakeNewScript = makeNewScript
XTension.addScriptHandler( "listUnits", listUnits)

# XTension.debugMode = True

XTension.startup()

if not XTension.ready:
	print( "Failed to make connection to XTension. Shutting down")
	sys.exit()


XTension.writeLog( "XTension Alexa hue Bridge Plugin v%s starting up" % (pluginVersion))

resetStatus = cResetStatus()

# startup the bridge
bridge = hueBridge()
bridge.startup()

XTension.writeLog( "v%s ready on interface %s:%s" % (pluginVersion, bridge.serverIp, bridge.serverPort))


